@@ -2,26 +2,59 @@
 // SPDX-License-Identifier: BSD-3-Clause
 // https://github.com/AcademySoftwareFoundation/OpenShadingLanguage
 
-#include <OpenImageIO/filesystem.h>
-#include <OpenImageIO/parallel.h>
 
-#include <pugixml.hpp>
+#ifndef __CUDACC__
+#    include <OpenImageIO/filesystem.h>
+#    include <OpenImageIO/parallel.h>
+#    include <OpenImageIO/timer.h>
+#    include <pugixml.hpp>
 
-#ifdef USING_OIIO_PUGI
+#    ifdef USING_OIIO_PUGI
 namespace pugi = OIIO::pugi;
+#    endif
+
+#    include <OSL/hashes.h>
+#    include "raytracer.h"
+#    include "shading.h"
+#    include "simpleraytracer.h"
 #endif
 
-#include "raytracer.h"
-#include "shading.h"
-#include "simpleraytracer.h"
+#include <cmath>
+
+// Create ustrings for all strings used by the free function renderer services.
+// Required to allow the reverse mapping of hash->string to work when processing messages
+namespace RS {
+namespace Strings {
+#define RS_STRDECL(str, var_name) const OSL::ustring var_name { str };
+#include "rs_strdecls.h"
+#undef RS_STRDECL
+}  // namespace Strings
+}  // namespace RS
+
+namespace RS {
+namespace {
+namespace Hashes {
+#define RS_STRDECL(str, var_name) \
+    constexpr OSL::ustringhash var_name(OSL::strhash(str));
+#include "rs_strdecls.h"
+#undef RS_STRDECL
+};  //namespace Hashes
+}  // unnamed namespace
+};  //namespace RS
+
 using namespace OSL;
 
-OSL_NAMESPACE_ENTER
+OSL_NAMESPACE_BEGIN
 
-static ustring u_camera("camera"), u_screen("screen");
-static ustring u_NDC("NDC"), u_raster("raster");
-static ustring u_perspective("perspective");
-static ustring u_s("s"), u_t("t");
+
+#ifndef __CUDACC__
+using ShaderGlobalsType = OSL::ShaderGlobals;
+#else
+using ShaderGlobalsType = OSL_CUDA::ShaderGlobals;
+#endif
+
+
+#ifndef __CUDACC__
 static TypeDesc TypeFloatArray2(TypeDesc::FLOAT, 2);
 static TypeDesc TypeFloatArray4(TypeDesc::FLOAT, 4);
 static TypeDesc TypeIntArray2(TypeDesc::INT, 2);
@@ -55,29 +88,29 @@ SimpleRaytracer::SimpleRaytracer()
 
     Matrix44 M;
     M.makeIdentity();
-    camera_params(M, u_perspective, 90.0f, 0.1f, 1000.0f, 256, 256);
+    camera_params(M, RS::Hashes::perspective, 90.0f, 0.1f, 1000.0f, 256, 256);
 
     // Set up getters
-    m_attr_getters[ustring("osl:version")] = &SimpleRaytracer::get_osl_version;
-    m_attr_getters[ustring("camera:resolution")]
+    m_attr_getters[RS::Hashes::osl_version] = &SimpleRaytracer::get_osl_version;
+    m_attr_getters[RS::Hashes::camera_resolution]
         = &SimpleRaytracer::get_camera_resolution;
-    m_attr_getters[ustring("camera:projection")]
+    m_attr_getters[RS::Hashes::camera_projection]
         = &SimpleRaytracer::get_camera_projection;
-    m_attr_getters[ustring("camera:pixelaspect")]
+    m_attr_getters[RS::Hashes::camera_pixelaspect]
         = &SimpleRaytracer::get_camera_pixelaspect;
-    m_attr_getters[ustring("camera:screen_window")]
+    m_attr_getters[RS::Hashes::camera_screen_window]
         = &SimpleRaytracer::get_camera_screen_window;
-    m_attr_getters[ustring("camera:fov")]  = &SimpleRaytracer::get_camera_fov;
-    m_attr_getters[ustring("camera:clip")] = &SimpleRaytracer::get_camera_clip;
-    m_attr_getters[ustring("camera:clip_near")]
+    m_attr_getters[RS::Hashes::camera_fov]  = &SimpleRaytracer::get_camera_fov;
+    m_attr_getters[RS::Hashes::camera_clip] = &SimpleRaytracer::get_camera_clip;
+    m_attr_getters[RS::Hashes::camera_clip_near]
         = &SimpleRaytracer::get_camera_clip_near;
-    m_attr_getters[ustring("camera:clip_far")]
+    m_attr_getters[RS::Hashes::camera_clip_far]
         = &SimpleRaytracer::get_camera_clip_far;
-    m_attr_getters[ustring("camera:shutter")]
+    m_attr_getters[RS::Hashes::camera_shutter]
         = &SimpleRaytracer::get_camera_shutter;
-    m_attr_getters[ustring("camera:shutter_open")]
+    m_attr_getters[RS::Hashes::camera_shutter_open]
         = &SimpleRaytracer::get_camera_shutter_open;
-    m_attr_getters[ustring("camera:shutter_close")]
+    m_attr_getters[RS::Hashes::camera_shutter_close]
         = &SimpleRaytracer::get_camera_shutter_close;
 }
 
@@ -125,7 +158,7 @@ SimpleRaytracer::attribute(string_view name, TypeDesc type, const void* value)
 
 void
 SimpleRaytracer::camera_params(const Matrix44& world_to_camera,
-                               ustring projection, float hfov, float hither,
+                               ustringhash projection, float hfov, float hither,
                                float yon, int xres, int yres)
 {
     m_world_to_camera  = world_to_camera;
@@ -166,7 +199,6 @@ strtobool(const char* str)
            || strcmp(str, "yes") == 0;
 }
 
-
 template<int N> struct ParamStorage {
     ParamStorage() : fparamindex(0), iparamindex(0), sparamindex(0) {}
 
@@ -248,6 +280,7 @@ parse_prefix_and_ints(string_view str, string_view prefix, int nvals, int* vals)
 void
 SimpleRaytracer::parse_scene_xml(const std::string& scenefile)
 {
+    ShaderMap shadermap;
     pugi::xml_document doc;
     pugi::xml_parse_result parse_result;
     if (OIIO::Strutil::ends_with(scenefile, ".xml")
@@ -304,17 +337,21 @@ SimpleRaytracer::parse_scene_xml(const std::string& scenefile)
             // load sphere
             pugi::xml_attribute center_attr = node.attribute("center");
             pugi::xml_attribute radius_attr = node.attribute("radius");
+
             if (center_attr && radius_attr) {
                 Vec3 center  = strtovec(center_attr.value());
                 float radius = OIIO::Strutil::from_string<float>(
                     radius_attr.value());
                 if (radius > 0) {
-                    pugi::xml_attribute light_attr = node.attribute("is_light");
-                    bool is_light = light_attr ? strtobool(light_attr.value())
-                                               : false;
-                    scene.add_sphere(Sphere(center, radius,
-                                            int(shaders().size()) - 1,
-                                            is_light));
+                    pugi::xml_attribute resolution_attr = node.attribute(
+                        "resolution");
+                    int resolution = 64;
+                    if (resolution_attr) {
+                        OIIO::string_view str = resolution_attr.value();
+                        OIIO::Strutil::parse_int(str, resolution);
+                    }
+                    scene.add_sphere(center, radius, int(shaders().size()) - 1,
+                                     resolution);
                 }
             }
         } else if (strcmp(node.name(), "Quad") == 0) {
@@ -323,14 +360,40 @@ SimpleRaytracer::parse_scene_xml(const std::string& scenefile)
             pugi::xml_attribute edge_x_attr = node.attribute("edge_x");
             pugi::xml_attribute edge_y_attr = node.attribute("edge_y");
             if (corner_attr && edge_x_attr && edge_y_attr) {
-                pugi::xml_attribute light_attr = node.attribute("is_light");
-                bool is_light = light_attr ? strtobool(light_attr.value())
-                                           : false;
-                Vec3 co       = strtovec(corner_attr.value());
-                Vec3 ex       = strtovec(edge_x_attr.value());
-                Vec3 ey       = strtovec(edge_y_attr.value());
-                scene.add_quad(
-                    Quad(co, ex, ey, int(shaders().size()) - 1, is_light));
+                Vec3 co = strtovec(corner_attr.value());
+                Vec3 ex = strtovec(edge_x_attr.value());
+                Vec3 ey = strtovec(edge_y_attr.value());
+
+                int resolution                      = 1;
+                pugi::xml_attribute resolution_attr = node.attribute(
+                    "resolution");
+                if (resolution_attr) {
+                    OIIO::string_view str = resolution_attr.value();
+                    OIIO::Strutil::parse_int(str, resolution);
+                }
+
+                scene.add_quad(co, ex, ey, int(shaders().size()) - 1,
+                               resolution);
+            }
+        } else if (strcmp(node.name(), "Model") == 0) {
+            // load .obj model
+            pugi::xml_attribute filename_attr = node.attribute("filename");
+            if (filename_attr) {
+                std::string filename = filename_attr.value();
+                std::vector<std::string> searchpath;
+                searchpath.emplace_back(
+                    OIIO::Filesystem::parent_path(scenefile));
+                std::string actual_filename
+                    = OIIO::Filesystem::searchpath_find(filename, searchpath,
+                                                        false);
+                if (actual_filename.empty()) {
+                    errhandler().errorfmt("Unable to find model file {}",
+                                          filename);
+                } else {
+                    // we got a valid filename, try to load the model
+                    scene.add_model(actual_filename, shadermap,
+                                    int(shaders().size() - 1), errhandler());
+                }
             }
         } else if (strcmp(node.name(), "Background") == 0) {
             pugi::xml_attribute res_attr = node.attribute("resolution");
@@ -340,7 +403,8 @@ SimpleRaytracer::parse_scene_xml(const std::string& scenefile)
             backgroundShaderID = int(shaders().size()) - 1;
         } else if (strcmp(node.name(), "ShaderGroup") == 0) {
             ShaderGroupRef group;
-            pugi::xml_attribute name_attr = node.attribute("name");
+            pugi::xml_attribute is_light_attr = node.attribute("is_light");
+            pugi::xml_attribute name_attr     = node.attribute("name");
             std::string name = name_attr ? name_attr.value() : "group";
             pugi::xml_attribute type_attr = node.attribute("type");
             std::string shadertype = type_attr ? type_attr.value() : "surface";
@@ -407,7 +471,29 @@ SimpleRaytracer::parse_scene_xml(const std::string& scenefile)
                 }
             }
             shadingsys->ShaderGroupEnd(*group);
-            shaders().push_back(group);
+            if (name_attr) {
+                if (auto it = shadermap.find(name); it != shadermap.end()) {
+                    int shaderID = it->second;
+                    if (shaderID >= 0 && shaderID < int(shaders().size())) {
+                        fprintf(stderr, "Updating shader %d - %s\n", shaderID,
+                                shadertype.c_str());
+                        // we already have a material under this name,
+                        Material& m = shaders()[shaderID];
+                        // TODO: could we query the shadertype directly from the ShaderGroup?
+                        if (shadertype == "displacement")
+                            m.disp = group;
+                        else if (shadertype == "surface")
+                            m.surf = group;
+                        // skip the rest which would add a new material
+                        continue;
+                    }
+                } else {
+                    shadermap.emplace(name, int(shaders().size()));
+                }
+            }
+            shaders().emplace_back(Material { group, nullptr });
+            m_shader_is_light.emplace_back(
+                is_light_attr ? strtobool(is_light_attr.value()) : false);
         } else {
             // unknown element?
         }
@@ -417,6 +503,8 @@ SimpleRaytracer::parse_scene_xml(const std::string& scenefile)
             "Error reading {}: Found multiple top-level elements", scenefile);
     if (shaders().empty())
         errhandler().severefmt("No shaders in scene");
+    if (scene.num_prims() == 0)
+        errhandler().severefmt("No primitives in scene");
     camera.finalize();
 }
 
@@ -482,12 +570,13 @@ bool
 SimpleRaytracer::get_inverse_matrix(ShaderGlobals* /*sg*/, Matrix44& result,
                                     ustringhash to, float /*time*/)
 {
-    if (to == u_camera || to == u_screen || to == u_NDC || to == u_raster) {
+    if (to == OSL::Hashes::camera || to == OSL::Hashes::screen
+        || to == OSL::Hashes::NDC || to == RS::Hashes::raster) {
         // clang-format off
         Matrix44 M = m_world_to_camera;
-        if (to == u_screen || to == u_NDC || to == u_raster) {
+        if (to == OSL::Hashes::screen || to == OSL::Hashes::NDC || to == RS::Hashes::raster) {
             float depthrange = (double)m_yon-(double)m_hither;
-            if (m_projection == u_perspective) {
+            if (m_projection == RS::Hashes::perspective) {
                 float tanhalffov = tanf (0.5f * m_fov * M_PI/180.0);
                 Matrix44 camera_to_screen (1/tanhalffov, 0, 0, 0,
                                            0, 1/tanhalffov, 0, 0,
@@ -501,7 +590,7 @@ SimpleRaytracer::get_inverse_matrix(ShaderGlobals* /*sg*/, Matrix44& result,
                                            0, 0, -m_hither/depthrange, 1);
                 M = M * camera_to_screen;
             }
-            if (to == u_NDC || to == u_raster) {
+            if (to == OSL::Hashes::NDC || to == RS::Hashes::raster) {
                 float screenleft = -1.0, screenwidth = 2.0;
                 float screenbottom = -1.0, screenheight = 2.0;
                 Matrix44 screen_to_ndc (1/screenwidth, 0, 0, 0,
@@ -509,7 +598,7 @@ SimpleRaytracer::get_inverse_matrix(ShaderGlobals* /*sg*/, Matrix44& result,
                                         0, 0, 1, 0,
                                         -screenleft/screenwidth, -screenbottom/screenheight, 0, 1);
                 M = M * screen_to_ndc;
-                if (to == u_raster) {
+                if (to == RS::Hashes::raster) {
                     Matrix44 ndc_to_raster (camera.xres, 0, 0, 0,
                                             0, camera.yres, 0, 0,
                                             0, 0, 1, 0,
@@ -584,7 +673,7 @@ SimpleRaytracer::get_userdata(bool derivatives, ustringhash name, TypeDesc type,
     // look up something specific to the primitive, rather than have hard-
     // coded names.
 
-    if (name == u_s && type == TypeFloat) {
+    if (name == RS::Hashes::s && type == TypeFloat) {
         ((float*)val)[0] = sg->u;
         if (derivatives) {
             ((float*)val)[1] = sg->dudx;
@@ -592,7 +681,7 @@ SimpleRaytracer::get_userdata(bool derivatives, ustringhash name, TypeDesc type,
         }
         return true;
     }
-    if (name == u_t && type == TypeFloat) {
+    if (name == RS::Hashes::t && type == TypeFloat) {
         ((float*)val)[0] = sg->v;
         if (derivatives) {
             ((float*)val)[1] = sg->dvdx;
@@ -638,7 +727,7 @@ SimpleRaytracer::get_camera_projection(ShaderGlobals* /*sg*/, bool /*derivs*/,
                                        ustringhash /*name*/, void* val)
 {
     if (type == TypeString) {
-        ((ustring*)val)[0] = m_projection;
+        ((ustringhash*)val)[0] = m_projection;
         return true;
     }
     return false;
@@ -786,34 +875,46 @@ SimpleRaytracer::get_camera_screen_window(ShaderGlobals* /*sg*/, bool derivs,
     }
     return false;
 }
+#endif  // #ifndef __CUDACC__
 
 
 
-void
-SimpleRaytracer::globals_from_hit(ShaderGlobals& sg, const Ray& r,
-                                  const Dual2<float>& t, int id)
+void OSL_HOSTDEVICE
+SimpleRaytracer::globals_from_hit(ShaderGlobalsType& sg, const Ray& r,
+                                  const Dual2<float>& t, int id, float u,
+                                  float v)
 {
+#ifndef __CUDACC__
     memset((char*)&sg, 0, sizeof(ShaderGlobals));
+#endif
+
+#ifndef __CUDACC__
+    const int meshid = std::upper_bound(scene.last_index.begin(),
+                                        scene.last_index.end(), id)
+                       - scene.last_index.begin();
+#else
+    const int meshid = m_meshids[id];
+#endif
+
     Dual2<Vec3> P = r.point(t);
     // We are missing the projection onto the surface here
-    sg.P          = P.val();
-    sg.dPdx       = P.dx();
-    sg.dPdy       = P.dy();
-    Dual2<Vec3> N = scene.normal(P, id);
-    sg.Ng = sg.N          = N.val();
-    Dual2<Vec2> uv        = scene.uv(P, N, sg.dPdu, sg.dPdv, id);
+    sg.P                  = P.val();
+    sg.dPdx               = P.dx();
+    sg.dPdy               = P.dy();
+    sg.N                  = scene.normal(P, sg.Ng, id, u, v);
+    Dual2<Vec2> uv        = scene.uv(P, sg.N, sg.dPdu, sg.dPdv, id, u, v);
     sg.u                  = uv.val().x;
     sg.dudx               = uv.dx().x;
     sg.dudy               = uv.dy().x;
     sg.v                  = uv.val().y;
     sg.dvdx               = uv.dx().y;
     sg.dvdy               = uv.dy().y;
-    sg.surfacearea        = scene.surfacearea(id);
+    sg.surfacearea        = m_mesh_surfacearea[meshid];
     Dual2<Vec3> direction = r.dual_direction();
     sg.I                  = direction.val();
     sg.dIdx               = direction.dx();
     sg.dIdy               = direction.dy();
-    sg.backfacing         = sg.N.dot(sg.I) > 0;
+    sg.backfacing         = sg.Ng.dot(sg.I) > 0;
     if (sg.backfacing) {
         sg.N  = -sg.N;
         sg.Ng = -sg.Ng;
@@ -821,41 +922,58 @@ SimpleRaytracer::globals_from_hit(ShaderGlobals& sg, const Ray& r,
     sg.raytype        = r.raytype;
     sg.flipHandedness = sg.dPdx.cross(sg.dPdy).dot(sg.N) < 0;
 
+#ifndef __CUDACC__
     // In our SimpleRaytracer, the "renderstate" itself just a pointer to
     // the ShaderGlobals.
     sg.renderstate = &sg;
+#endif
 }
 
-Vec3
+
+
+OSL_HOSTDEVICE Vec3
 SimpleRaytracer::eval_background(const Dual2<Vec3>& dir, ShadingContext* ctx,
                                  int bounce)
 {
-    ShaderGlobals sg;
+    ShaderGlobalsType sg;
     memset((char*)&sg, 0, sizeof(ShaderGlobals));
     sg.I    = dir.val();
     sg.dIdx = dir.dx();
     sg.dIdy = dir.dy();
     if (bounce >= 0)
         sg.raytype = bounce > 0 ? Ray::DIFFUSE : Ray::CAMERA;
-    shadingsys->execute(*ctx, *m_shaders[backgroundShaderID], sg);
-    return process_background_closure(sg.Ci);
+#ifndef __CUDACC__
+    shadingsys->execute(*ctx, *m_shaders[backgroundShaderID].surf, sg);
+#else
+    alignas(8) char closure_pool[256];
+    execute_shader(sg, render_params.bg_id, closure_pool);
+#endif
+    return process_background_closure((const ClosureColor*)sg.Ci);
 }
 
 Color3
 SimpleRaytracer::subpixel_radiance(float x, float y, Sampler& sampler,
                                    ShadingContext* ctx)
 {
-    Ray r = camera.get(x, y);
+#ifdef __CUDACC__
+    // Scratch space for the output closures
+    alignas(8) char closure_pool[256];
+    alignas(8) char light_closure_pool[256];
+#endif
+
+    constexpr float inf = std::numeric_limits<float>::infinity();
+    Ray r               = camera.get(x, y);
     Color3 path_weight(1, 1, 1);
     Color3 path_radiance(0, 0, 0);
     int prev_id    = -1;
-    float bsdf_pdf = std::numeric_limits<
-        float>::infinity();  // camera ray has only one possible direction
+    float bsdf_pdf = inf;  // camera ray has only one possible direction
+
     for (int b = 0; b <= max_bounces; b++) {
+        ShaderGlobalsType sg;
+
         // trace the ray against the scene
-        Dual2<float> t;
-        int id = prev_id;
-        if (!scene.intersect(r, t, id)) {
+        Intersection hit = scene.intersect(r, inf, prev_id);
+        if (hit.t == inf) {
             // we hit nothing? check background shader
             if (backgroundShaderID >= 0) {
                 if (b > 0 && backgroundResolution > 0) {
@@ -875,24 +993,56 @@ SimpleRaytracer::subpixel_radiance(float x, float y, Sampler& sampler,
         }
 
         // construct a shader globals for the hit point
-        ShaderGlobals sg;
-        globals_from_hit(sg, r, t, id);
-        const float radius = r.radius + r.spread * t.val();
-        int shaderID       = scene.shaderid(id);
-        if (shaderID < 0 || !m_shaders[shaderID])
+        globals_from_hit(sg, r, hit.t, hit.id, hit.u, hit.v);
+
+        if (show_globals) {
+            // visualize the main fields of the shader globals
+            Vec3 v = sg.Ng;
+            if (show_globals == 2)
+                v = sg.N;
+            if (show_globals == 3)
+                v = sg.dPdu.normalize();
+            if (show_globals == 4)
+                v = sg.dPdv.normalize();
+            if (show_globals == 5)
+                v = Vec3(sg.u, sg.v, 0);
+            Color3 c(v.x, v.y, v.z);
+            if (show_globals != 5)
+                c = c * 0.5f + Color3(0.5f);
+            path_radiance += path_weight * c;
+            break;
+        }
+
+        const float radius = r.radius + r.spread * hit.t;
+
+        int shaderID = scene.shaderid(hit.id);
+
+#ifndef __CUDACC__
+        if (shaderID < 0 || !m_shaders[shaderID].surf)
             break;  // no shader attached? done
 
         // execute shader and process the resulting list of closures
-        shadingsys->execute(*ctx, *m_shaders[shaderID], sg);
+        shadingsys->execute(*ctx, *m_shaders[shaderID].surf, sg);
+#else
+        if (shaderID < 0)
+            break;  // no shader attached? done
+        execute_shader(sg, shaderID, closure_pool);
+#endif
         ShadingResult result;
         bool last_bounce = b == max_bounces;
-        process_closure(sg, result, sg.Ci, last_bounce);
+        process_closure(sg, result, (const ClosureColor*)sg.Ci, last_bounce);
+
+#ifndef __CUDACC__
+        const size_t lightprims_size = m_lightprims.size();
+#endif
 
         // add self-emission
         float k = 1;
-        if (scene.islight(id)) {
+        if (m_shader_is_light[shaderID] && lightprims_size > 0) {
+            const float light_pick_pdf = 1.0f / lightprims_size;
             // figure out the probability of reaching this point
-            float light_pdf = scene.shapepdf(id, r.origin, sg.P);
+            float light_pdf = light_pick_pdf
+                              * scene.shapepdf(hit.id, r.origin, sg.P);
             k = MIS::power_heuristic<MIS::WEIGHT_EVAL>(bsdf_pdf, light_pdf);
         }
         path_radiance += path_weight * k * result.Le;
@@ -929,49 +1079,72 @@ SimpleRaytracer::subpixel_radiance(float x, float y, Sampler& sampler,
                              * MIS::power_heuristic<MIS::WEIGHT_WEIGHT>(bg_pdf,
                                                                         b.pdf);
             if ((contrib.x + contrib.y + contrib.z) > 0) {
-                int shadow_id  = id;
-                Ray shadow_ray = Ray(sg.P, bg_dir.val(), radius, 0,
-                                     Ray::SHADOW);
-                Dual2<float> shadow_dist;
-                if (!scene.intersect(shadow_ray, shadow_dist,
-                                     shadow_id))  // ray reached the background?
+                ShaderGlobalsType shadow_sg;
+                Ray shadow_ray          = Ray(sg.P, bg_dir.val(), radius, 0,
+                                              Ray::SHADOW);
+                Intersection shadow_hit = scene.intersect(shadow_ray, inf,
+                                                          hit.id);
+                if (shadow_hit.t == inf)  // ray reached the background?
                     path_radiance += contrib;
             }
         }
 
-        // trace one ray to each light
-        for (int lid = 0; lid < scene.num_prims(); lid++) {
-            if (lid == id)
-                continue;  // skip self
-            if (!scene.islight(lid))
-                continue;  // doesn't want to be sampled as a light
-            int shaderID = scene.shaderid(lid);
-            if (shaderID < 0 || !m_shaders[shaderID])
-                continue;  // no shader attached to this light
-            // sample a random direction towards the object
-            float light_pdf;
-            Vec3 ldir      = scene.sample(lid, sg.P, xi, yi, light_pdf);
-            BSDF::Sample b = result.bsdf.eval(-sg.I, ldir);
-            Color3 contrib = path_weight * b.weight
-                             * MIS::power_heuristic<MIS::EVAL_WEIGHT>(light_pdf,
-                                                                      b.pdf);
-            if ((contrib.x + contrib.y + contrib.z) > 0) {
-                Ray shadow_ray = Ray(sg.P, ldir, radius, 0, Ray::SHADOW);
-                // trace a shadow ray and see if we actually hit the target
-                // in this tiny renderer, tracing a ray is probably cheaper than evaluating the light shader
-                int shadow_id = id;  // ignore self hit
-                Dual2<float> shadow_dist;
-                if (scene.intersect(shadow_ray, shadow_dist, shadow_id)
-                    && shadow_id == lid) {
-                    // setup a shader global for the point on the light
-                    ShaderGlobals light_sg;
-                    globals_from_hit(light_sg, shadow_ray, shadow_dist, lid);
-                    // execute the light shader (for emissive closures only)
-                    shadingsys->execute(*ctx, *m_shaders[shaderID], light_sg);
-                    ShadingResult light_result;
-                    process_closure(light_sg, light_result, light_sg.Ci, true);
-                    // accumulate contribution
-                    path_radiance += contrib * light_result.Le;
+        // trace a shadow ray to one of the light emitting primitives
+        if (lightprims_size > 0) {
+            const float light_pick_pdf = 1.0f / lightprims_size;
+
+            // uniform probability for each light
+            float xl = xi * lightprims_size;
+            int ls   = floorf(xl);
+            xl -= ls;
+
+            uint32_t lid = m_lightprims[ls];
+            if (lid != hit.id) {
+                int shaderID = scene.shaderid(lid);
+
+                // sample a random direction towards the object
+                LightSample sample = scene.sample(lid, sg.P, xl, yi);
+                BSDF::Sample b     = result.bsdf.eval(-sg.I, sample.dir);
+                Color3 contrib     = path_weight * b.weight
+                                 * MIS::power_heuristic<MIS::EVAL_WEIGHT>(
+                                     light_pick_pdf * sample.pdf, b.pdf);
+                if ((contrib.x + contrib.y + contrib.z) > 0) {
+                    ShaderGlobalsType light_sg;
+                    Ray shadow_ray = Ray(sg.P, sample.dir, radius, 0,
+                                         Ray::SHADOW);
+                    // trace a shadow ray and see if we actually hit the target
+                    // in this tiny renderer, tracing a ray is probably cheaper than evaluating the light shader
+                    Intersection shadow_hit
+                        = scene.intersect(shadow_ray, sample.dist, hit.id, lid);
+
+#ifndef __CUDACC__
+                    const bool did_hit = shadow_hit.t == sample.dist;
+#else
+                    // The hit distance on the device is not as precise as on
+                    // the CPU, so we need to allow a little wiggle room. An
+                    // epsilon of 1e-3f empirically gives results that closely
+                    // match the CPU for the test scenes, so that's what we're
+                    // using.
+                    const bool did_hit = fabsf(shadow_hit.t - sample.dist)
+                                         < 1e-3f;
+#endif
+                    if (did_hit) {
+                        // setup a shader global for the point on the light
+                        globals_from_hit(light_sg, shadow_ray, sample.dist, lid,
+                                         sample.u, sample.v);
+#ifndef __CUDACC__
+                        // execute the light shader (for emissive closures only)
+                        shadingsys->execute(*ctx, *m_shaders[shaderID].surf,
+                                            light_sg);
+#else
+                        execute_shader(light_sg, shaderID, light_closure_pool);
+#endif
+                        ShadingResult light_result;
+                        process_closure(light_sg, light_result,
+                                        (const ClosureColor*)light_sg.Ci, true);
+                        // accumulate contribution
+                        path_radiance += contrib * light_result.Le;
+                    }
                 }
             }
         }
@@ -988,20 +1161,21 @@ SimpleRaytracer::subpixel_radiance(float x, float y, Sampler& sampler,
         if (!(path_weight.x > 0) && !(path_weight.y > 0)
             && !(path_weight.z > 0))
             break;  // filter out all 0's or NaNs
-        prev_id  = id;
+        prev_id  = hit.id;
         r.origin = sg.P;
     }
     return path_radiance;
 }
 
-Color3
+
+OSL_HOSTDEVICE Color3
 SimpleRaytracer::antialias_pixel(int x, int y, ShadingContext* ctx)
 {
     Color3 result(0, 0, 0);
     for (int si = 0, n = aa * aa; si < n; si++) {
         Sampler sampler(x, y, si);
         // jitter pixel coordinate [0,1)^2
-        Vec3 j = sampler.get();
+        Vec3 j = no_jitter ? Vec3(0.5f, 0.5f, 0) : sampler.get();
         // warp distribution to approximate a tent filter [-1,+1)^2
         j.x *= 2;
         j.x = j.x < 1 ? sqrtf(j.x) - 1 : 1 - sqrtf(2 - j.x);
@@ -1017,14 +1191,17 @@ SimpleRaytracer::antialias_pixel(int x, int y, ShadingContext* ctx)
 }
 
 
+#ifndef __CUDACC__
 void
 SimpleRaytracer::prepare_render()
 {
     // Retrieve and validate options
     aa                = std::max(1, options.get_int("aa"));
+    no_jitter         = options.get_int("no_jitter") != 0;
     max_bounces       = options.get_int("max_bounces");
     rr_depth          = options.get_int("rr_depth");
     show_albedo_scale = options.get_float("show_albedo_scale");
+    show_globals      = options.get_int("show_globals");
 
     // prepare background importance table (if requested)
     if (backgroundResolution > 0 && backgroundShaderID >= 0) {
@@ -1045,13 +1222,184 @@ SimpleRaytracer::prepare_render()
         // we aren't directly evaluating the background
         backgroundResolution = 0;
     }
+
+    prepare_geometry();
+
+    // build bvh and prepare triangles
+    scene.prepare(errhandler());
+    prepare_lights();
+
+#    if 0
+    // dump scene to disk as obj for debugging purposes
+    // TODO: make this a feature?
+    FILE* fp = fopen("/tmp/test.obj", "w");
+    for (Vec3 v : scene.verts)
+        fprintf(fp, "v %.9g %.9g %.9g\n", v.x, v.y, v.z);
+    for (TriangleIndices t : scene.triangles)
+        fprintf(fp, "f %d %d %d\n", 1 + t.a, 1 + t.b, 1 + t.c);
+    fclose(fp);
+#    endif
 }
 
 
 
+void
+SimpleRaytracer::prepare_lights()
+{
+    m_mesh_surfacearea.reserve(scene.last_index.size());
+
+    // measure the total surface area of each mesh
+    int first_index = 0;
+    for (int last_index : scene.last_index) {
+        float area = 0;
+        for (int index = first_index; index < last_index; index++) {
+            area += scene.primitivearea(index);
+        }
+        m_mesh_surfacearea.emplace_back(area);
+        first_index = last_index;
+    }
+    // collect all light emitting triangles
+    for (unsigned t = 0, n = scene.num_prims(); t < n; t++) {
+        int shaderID = scene.shaderid(t);
+        if (shaderID < 0 || !m_shaders[shaderID].surf)
+            continue;  // no shader attached
+        if (m_shader_is_light[shaderID])
+            m_lightprims.emplace_back(t);
+    }
+    if (!m_lightprims.empty())
+        errhandler().infofmt("Found {} triangles to be treated as lights",
+                             m_lightprims.size());
+}
+
+
+void
+SimpleRaytracer::prepare_geometry()
+{
+    bool have_displacement = false;
+    for (const Material& m : shaders()) {
+        if (m.disp) {
+            have_displacement = true;
+            break;
+        }
+    }
+    if (have_displacement) {
+        errhandler().infofmt("Evaluating displacement shaders");
+        // Loop through all triangles and run displacement shader if there is one
+        // or copy the input point if there is none
+        std::vector<Vec3> disp_verts(scene.verts.size(), Vec3(0, 0, 0));
+        std::vector<int> valance(
+            scene.verts.size(),
+            0);  // number of times each vertex has been displaced
+
+        OSL::PerThreadInfo* thread_info = shadingsys->create_thread_info();
+        ShadingContext* ctx             = shadingsys->get_context(thread_info);
+
+        bool has_smooth_normals = false;
+        for (int primID = 0, nprims = scene.triangles.size(); primID < nprims;
+             primID++) {
+            Vec3 p[3], n[3];
+            Vec2 uv[3];
+
+            p[0] = scene.verts[scene.triangles[primID].a];
+            p[1] = scene.verts[scene.triangles[primID].b];
+            p[2] = scene.verts[scene.triangles[primID].c];
+
+            valance[scene.triangles[primID].a]++;
+            valance[scene.triangles[primID].b]++;
+            valance[scene.triangles[primID].c]++;
+
+            int shaderID = scene.shaderid(primID);
+            if (shaderID < 0 || !m_shaders[shaderID].disp) {
+                disp_verts[scene.triangles[primID].a] += p[0];
+                disp_verts[scene.triangles[primID].b] += p[1];
+                disp_verts[scene.triangles[primID].c] += p[2];
+                continue;
+            }
+
+
+            Vec3 Ng    = (p[0] - p[1]).cross(p[0] - p[2]);
+            float area = 0.5f * Ng.length();
+            Ng         = Ng.normalize();
+            if (scene.n_triangles[primID].a >= 0) {
+                n[0]               = scene.normals[scene.n_triangles[primID].a];
+                n[1]               = scene.normals[scene.n_triangles[primID].b];
+                n[2]               = scene.normals[scene.n_triangles[primID].c];
+                has_smooth_normals = true;
+            } else {
+                n[0] = n[1] = n[2] = Ng;
+            }
+
+            if (scene.uv_triangles[primID].a >= 0) {
+                uv[0] = scene.uvs[scene.uv_triangles[primID].a];
+                uv[1] = scene.uvs[scene.uv_triangles[primID].b];
+                uv[2] = scene.uvs[scene.uv_triangles[primID].c];
+            } else {
+                uv[0] = uv[1] = uv[2] = Vec2(0, 0);
+            }
+
+            // displace each vertex
+            for (int i = 0; i < 3; i++) {
+                ShaderGlobals sg = {};
+                sg.P             = p[i];
+                sg.Ng            = Ng;
+                sg.N             = n[i];
+                sg.u             = uv[i].x;
+                sg.v             = uv[i].y;
+                sg.I             = (p[i] - camera.eye).normalize();
+                sg.surfacearea   = area;
+                sg.renderstate   = &sg;
+
+                shadingsys->execute(*ctx, *m_shaders[shaderID].disp, sg);
+
+                p[i] = sg.P;
+            }
+            disp_verts[scene.triangles[primID].a] += p[0];
+            disp_verts[scene.triangles[primID].b] += p[1];
+            disp_verts[scene.triangles[primID].c] += p[2];
+        }
+
+        // release context
+        shadingsys->release_context(ctx);
+        shadingsys->destroy_thread_info(thread_info);
+
+        // average each vertex by the number of times it was displaced
+        for (int i = 0, n = scene.verts.size(); i < n; i++) {
+            if (valance[i] > 0)
+                disp_verts[i] /= float(valance[i]);
+            else
+                disp_verts[i] = scene.verts[i];
+        }
+        // replace old data with the new
+        scene.verts = std::move(disp_verts);
+
+        if (has_smooth_normals) {
+            // Recompute the vertex normals (if we had some)
+            std::vector<Vec3> disp_normals(scene.normals.size(), Vec3(0, 0, 0));
+            for (int primID = 0, nprims = scene.triangles.size();
+                 primID < nprims; primID++) {
+                if (scene.n_triangles[primID].a >= 0) {
+                    Vec3 p[3];
+                    p[0] = scene.verts[scene.triangles[primID].a];
+                    p[1] = scene.verts[scene.triangles[primID].b];
+                    p[2] = scene.verts[scene.triangles[primID].c];
+                    // don't normalize to weight by area
+                    Vec3 Ng = (p[0] - p[1]).cross(p[0] - p[2]);
+                    disp_normals[scene.n_triangles[primID].a] += Ng;
+                    disp_normals[scene.n_triangles[primID].b] += Ng;
+                    disp_normals[scene.n_triangles[primID].c] += Ng;
+                }
+            }
+            for (Vec3& n : disp_normals)
+                n = n.normalize();
+            scene.normals = std::move(disp_normals);
+        }
+    }
+}
+
 void
 SimpleRaytracer::render(int xres, int yres)
 {
+    OIIO::Timer timer;
     ShadingSystem* shadingsys = this->shadingsys;
     OIIO::parallel_for_chunked(
         0, yres, 0, [&, this](int64_t ybegin, int64_t yend) {
@@ -1077,8 +1425,20 @@ SimpleRaytracer::render(int xres, int yres)
             shadingsys->release_context(ctx);
             shadingsys->destroy_thread_info(thread_info);
         });
+    double rendertime = timer();
+    errhandler().infofmt("Rendered {}x{} image with {} samples in {}", xres,
+                         yres, aa * aa,
+                         OIIO::Strutil::timeintervalformat(rendertime, 2));
 }
 
 
 
-OSL_NAMESPACE_EXIT
+void
+SimpleRaytracer::clear()
+{
+    shaders().clear();
+}
+
+#endif  // #ifndef __CUDACC__
+
+OSL_NAMESPACE_END
